"use strict";
const easingNames = ["linear", "easeInQuad", "easeOutQuad", "easeInOutQuad", "easeInCubic", "easeOutCubic", "easeInOutCubic", "easeInQuart", "easeOutQuart", "easeInOutQuart", "easeInQuint", "easeOutQuint", "easeInOutQuint", "easeInSine", "easeOutSine", "easeInOutSine", "easeInExpo", "easeOutExpo", "easeInOutExpo", "easeInCirc", "easeOutCirc", "easeInOutCirc", "easeOutBounce", "easeInBack", "easeOutBack", "easeInOutBack", "elastic", "swingFromTo", "swingFrom", "swingTo", "bounce", "bouncePast", "easeFromTo", "easeFrom", "easeTo"]

var EasingFunctions = {
    // no easing, no acceleration
    "linear": function(t)
    {
        return t
    },
    // accelerating from zero velocity
    "easeInQuad": function(t)
    {
        return t * t
    },
    // decelerating to zero velocity
    "easeOutQuad": function(t)
    {
        return t * (2 - t)
    },
    // acceleration until halfway, then deceleration
    "easeInOutQuad": function(t)
    {
        return t < .5 ? 2 * t * t : -1 + (4 - 2 * t) * t
    },
    // accelerating from zero velocity 
    "easeInCubic": function(t)
    {
        return t * t * t
    },
    // decelerating to zero velocity 
    "easeOutCubic": function(t)
    {
        return (--t) * t * t + 1
    },
    // acceleration until halfway, then deceleration 
    "easeInOutCubic": function(t)
    {
        return t < .5 ? 4 * t * t * t : (t - 1) * (2 * t - 2) * (2 * t - 2) + 1
    },
    // accelerating from zero velocity 
    "easeInQuart": function(t)
    {
        return t * t * t * t
    },
    // decelerating to zero velocity 
    "easeOutQuart": function(t)
    {
        return 1 - (--t) * t * t * t
    },
    // acceleration until halfway, then deceleration
    "easeInOutQuart": function(t)
    {
        return t < .5 ? 8 * t * t * t * t : 1 - 8 * (--t) * t * t * t
    },
    // accelerating from zero velocity
    "easeInQuint": function(t)
    {
        return t * t * t * t * t
    },
    // decelerating to zero velocity
    "easeOutQuint": function(t)
    {
        return 1 + (--t) * t * t * t * t
    },
    // acceleration until halfway, then deceleration 
    "easeInOutQuint": function(t)
    {
        return t < .5 ? 16 * t * t * t * t * t : 1 + 16 * (--t) * t * t * t * t
    },

    "easeInSine": function(pos)
    {
        return -Math.cos(pos * (Math.PI / 2)) + 1;
    },

    "easeOutSine": function(pos)
    {
        return Math.sin(pos * (Math.PI / 2));
    },

    "easeInOutSine": function(pos)
    {
        return (-0.5 * (Math.cos(Math.PI * pos) - 1));
    },

    "easeInExpo": function(pos)
    {
        return (pos === 0) ? 0 : Math.pow(2, 10 * (pos - 1));
    },

    "easeOutExpo": function(pos)
    {
        return (pos === 1) ? 1 : -Math.pow(2, -10 * pos) + 1;
    },

    "easeInOutExpo": function(pos)
    {
        if (pos === 0) return 0;
        if (pos === 1) return 1;
        if ((pos /= 0.5) < 1) return 0.5 * Math.pow(2, 10 * (pos - 1));
        return 0.5 * (-Math.pow(2, -10 * --pos) + 2);
    },

    "easeInCirc": function(pos)
    {
        return -(Math.sqrt(1 - (pos * pos)) - 1);
    },

    "easeOutCirc": function(pos)
    {
        return Math.sqrt(1 - Math.pow((pos - 1), 2));
    },

    "easeInOutCirc": function(pos)
    {
        if ((pos /= 0.5) < 1) return -0.5 * (Math.sqrt(1 - pos * pos) - 1);
        return 0.5 * (Math.sqrt(1 - (pos -= 2) * pos) + 1);
    },

    "easeOutBounce": function(pos)
    {
        if ((pos) < (1 / 2.75))
        {
            return (7.5625 * pos * pos);
        }
        else if (pos < (2 / 2.75))
        {
            return (7.5625 * (pos -= (1.5 / 2.75)) * pos + 0.75);
        }
        else if (pos < (2.5 / 2.75))
        {
            return (7.5625 * (pos -= (2.25 / 2.75)) * pos + 0.9375);
        }
        else
        {
            return (7.5625 * (pos -= (2.625 / 2.75)) * pos + 0.984375);
        }
    },

    "easeInBack": function(pos)
    {
        var s = 1.70158;
        return (pos) * pos * ((s + 1) * pos - s);
    },

    "easeOutBack": function(pos)
    {
        var s = 1.70158;
        return (pos = pos - 1) * pos * ((s + 1) * pos + s) + 1;
    },

    "easeInOutBack": function(pos)
    {
        var s = 1.70158;
        if ((pos /= 0.5) < 1) return 0.5 * (pos * pos * (((s *= (1.525)) + 1) * pos - s));
        return 0.5 * ((pos -= 2) * pos * (((s *= (1.525)) + 1) * pos + s) + 2);
    },

    "elastic": function(pos)
    {
        return -1 * Math.pow(4, -8 * pos) * Math.sin((pos * 6 - 1) * (2 * Math.PI) / 2) + 1;
    },

    "swingFromTo": function(pos)
    {
        var s = 1.70158;
        return ((pos /= 0.5) < 1) ? 0.5 * (pos * pos * (((s *= (1.525)) + 1) * pos - s)) :
            0.5 * ((pos -= 2) * pos * (((s *= (1.525)) + 1) * pos + s) + 2);
    },

    "swingFrom": function(pos)
    {
        var s = 1.70158;
        return pos * pos * ((s + 1) * pos - s);
    },

    "swingTo": function(pos)
    {
        var s = 1.70158;
        return (pos -= 1) * pos * ((s + 1) * pos + s) + 1;
    },

    "bounce": function(pos)
    {
        if (pos < (1 / 2.75))
        {
            return (7.5625 * pos * pos);
        }
        else if (pos < (2 / 2.75))
        {
            return (7.5625 * (pos -= (1.5 / 2.75)) * pos + 0.75);
        }
        else if (pos < (2.5 / 2.75))
        {
            return (7.5625 * (pos -= (2.25 / 2.75)) * pos + 0.9375);
        }
        else
        {
            return (7.5625 * (pos -= (2.625 / 2.75)) * pos + 0.984375);
        }
    },

    "bouncePast": function(pos)
    {
        if (pos < (1 / 2.75))
        {
            return (7.5625 * pos * pos);
        }
        else if (pos < (2 / 2.75))
        {
            return 2 - (7.5625 * (pos -= (1.5 / 2.75)) * pos + 0.75);
        }
        else if (pos < (2.5 / 2.75))
        {
            return 2 - (7.5625 * (pos -= (2.25 / 2.75)) * pos + 0.9375);
        }
        else
        {
            return 2 - (7.5625 * (pos -= (2.625 / 2.75)) * pos + 0.984375);
        }
    },

    "easeFromTo": function(pos)
    {
        if ((pos /= 0.5) < 1) return 0.5 * Math.pow(pos, 4);
        return -0.5 * ((pos -= 2) * Math.pow(pos, 3) - 2);
    },

    "easeFrom": function(pos)
    {
        return Math.pow(pos, 4);
    },

    "easeTo": function(pos)
    {
        return Math.pow(pos, 0.25);
    }
}

    function cos(x)
    {
        return Math.cos(x * Math.PI / 180);
    }

    function sin(x)
    {
        return Math.sin(x * Math.PI / 180);
    }

    function random(x)
    {
        return Math.random() * x;
    }

    function hslToRgb(hue, saturation, lightness)
    {
        // based on algorithm from http://en.wikipedia.org/wiki/HSL_and_HSV#Converting_to_RGB
        if (hue == undefined)
        {
            return [0, 0, 0];
        }

        var chroma = (1 - Math.abs((2 * lightness) - 1)) * saturation;
        var huePrime = hue / 60;
        var secondComponent = chroma * (1 - Math.abs((huePrime % 2) - 1));

        huePrime = Math.floor(huePrime);
        var red;
        var green;
        var blue;

        if (huePrime === 0)
        {
            red = chroma;
            green = secondComponent;
            blue = 0;
        }
        else if (huePrime === 1)
        {
            red = secondComponent;
            green = chroma;
            blue = 0;
        }
        else if (huePrime === 2)
        {
            red = 0;
            green = chroma;
            blue = secondComponent;
        }
        else if (huePrime === 3)
        {
            red = 0;
            green = secondComponent;
            blue = chroma;
        }
        else if (huePrime === 4)
        {
            red = secondComponent;
            green = 0;
            blue = chroma;
        }
        else if (huePrime === 5)
        {
            red = chroma;
            green = 0;
            blue = secondComponent;
        }

        var lightnessAdjustment = lightness - (chroma / 2);
        red += lightnessAdjustment;
        green += lightnessAdjustment;
        blue += lightnessAdjustment;

        return [Math.round(red * 255), Math.round(green * 255), Math.round(blue * 255)];

    };
var hsltorgb = hslToRgb;

function colorToHex(color)
{
    if (color.startsWith("#"))
    {
        return color
    }
    if (color.startsWith("hsl"))
    {
        let[h, s, l] = color.split('(')[1].split(')')[0].split(',').map(x => {
            return x.trim()
        })
        color = 'rgb(' + hslToRgb(h, s / 100, l / 100).join(',') + ')'
    }
    if (color.startsWith("rgb"))
    {
        let[r, g, b] = color.split('(')[1].split(')')[0].split(',').map(x => {
            x = parseInt(x.trim()).toString(16);
            return (x.length == 1) ? "0" + x : x;
        })
        return "#" + r + g + b
    }
}
var colortohex = colorToHex;

function lerpColor(a, b, x)
{
    return lerpHexColor(colorToHex(a), colorToHex(b), x)
}
var lerpcolor = lerpColor;

function lerpHexColor(a, b, amount)
{

    var ah = parseInt(a.replace(/#/g, ''), 16),
        ar = ah >> 16,
        ag = ah >> 8 & 0xff,
        ab = ah & 0xff,
        bh = parseInt(b.replace(/#/g, ''), 16),
        br = bh >> 16,
        bg = bh >> 8 & 0xff,
        bb = bh & 0xff,
        rr = ar + amount * (br - ar),
        rg = ag + amount * (bg - ag),
        rb = ab + amount * (bb - ab);

    return '#' + ((1 << 24) + (rr << 16) + (rg << 8) + rb | 0).toString(16).slice(1);
}
var lerphexcolor = lerpHexColor;

function lerpUnlerp(minOutput, maxOutput, minInput, maxInput, x, clamp = false)
{
    if (clamp)
    {
        if (x > maxInput) x = maxInput
        if (x < minInput) x = minInput
    }
    return minOutput + ((x - minInput) / (maxInput - minInput)) * (maxOutput - minOutput);
}
var lerpunlerp = lerpUnlerp;

function unlerp(min, max, x, clamp = false)
{
    if (clamp)
    {
        if (x > max) x = max
        if (x < min) x = min
    }
    if (min === max && x >= max) return 1
    return (x - min) / (max - min);
}

function lerp(min, max, x)
{
    let a = typeof min
    let b = typeof max
    if (a === b)
    {
        if (a === "number")
        {
            return x * (max - min) + min;
        }
        else
        {
            return lerpColor(min, max, x)
        }
    }
    else
    {
        if (a === "number")
        {
            max = parseFloat(max)
        }
        else
        {
            min = parseFloat(min)
        }
        return x * (max - min) + min;
    }
}

{
    C3.Behaviors.skymen_Skymen_SpritefontDX.Instance = class Skymen_SpritefontDXInstance extends C3.SDKBehaviorInstanceBase
    {
        constructor(behInst, properties)
        {
            super(behInst);

            this.TWData = {}
            this.TWTime = 0

            if (properties)
            {
                this.TWParams = properties[0];
                this.TWParamsOBJ = this.parseTypewriterParams(this.TWParams);
                this.TWEasing = easingNames[properties[1]];
            }
            this.curTypedWidth = ""
            this.curTypedHeight = ""
            this.text = ""
            this.lastKnown = ""
            this.typewriterPaused = false;
            this.typewriterActive = false;
            this.LastLetterID = 0;
            this.typePauseTime = 0;
            this.typewriterTagData = [];
            this.nextLetterTime = 0;
            this.typewriterWait = false;
            this.firstFrame = true;
            this.animated = false;
            this.SetTextCall = false;

            this.linkedDictionnaryUID = -1
            this.parsedText = []
            this.AnimFunctions = {}
            if (!this._inst._objectType._SFDXAliasFunctions) this._inst._objectType._SFDXAliasFunctions = {}

            this._StartTicking2()

            // Opt-in to getting calls to Tick()
            // this._StartTicking();
        }

        Release()
        {
            super.Release();
        }

        specialJSONstringify(obj)
        {
            var placeholder = '____PLACEHOLDER____';
            var fns = [];
            var json = JSON.stringify(obj, function(key, value)
            {
                if (typeof value === 'function')
                {
                    fns.push(value);
                    return placeholder;
                }
                return value;
            }, 2);
            json = json.replace(new RegExp('"' + placeholder + '"', 'g'), function(_)
            {
                return fns.shift();
            });
            return json;
        };

        specialJSONparse(str)
        {
            let val;
            eval('val =' + str + ';');
            return val;
        }

        SaveToJson()
        {
            return {
                curTypedWidth: this.curTypedWidth,
                curTypedHeight: this.curTypedHeight,
                text: this.text,
                typewriterPaused: this.typewriterPaused,
                typewriterActive: this.typewriterActive,
                LastLetterID: this.LastLetterID,
                typePauseTime: this.typePauseTime,
                nextLetterTime: this.nextLetterTime,
                typewriterWait: this.typewriterWait,
                firstFrame: this.firstFrame,
                animated: this.animated,
                SetTextCall: this.SetTextCall,
                lastKnown: this.lastKnown,
                TWTime: this.TWTime,
                TWEasing: this.TWEasing,

                typewriterTagData: JSON.stringify(this.typewriterTagData),
                TWData: JSON.stringify(this.TWData),
                TWParamsOBJ: JSON.stringify(this.TWParamsOBJ),
                TWParams: JSON.stringify(this.TWParams),

                AnimFunctions: this.specialJSONstringify(this.AnimFunctions),
                parsedText: this.specialJSONstringify(this.parsedText),
                _SFDXAliasFunctions: this.specialJSONstringify(this._inst._objectType._SFDXAliasFunctions),

                linkedDictionnaryUID: this.linkedDictionnaryUID
            };
        }

        LoadFromJson(o)
        {
            this.curTypedHeight = o.curTypedHeight
            this.curTypedWidth = o.curTypedWidth
            this.text = o.text
            this.typewriterPaused = o.typewriterPaused
            this.typewriterActive = o.typewriterActive
            this.LastLetterID = o.LastLetterID
            this.typePauseTime = o.typePauseTime
            this.nextLetterTime = o.nextLetterTime
            this.typewriterWait = o.typewriterWait
            this.firstFrame = o.firstFrame
            this.animated = o.animated
            this.SetTextCall = o.SetTextCall
            this.lastKnown = o.lastKnown
            this.TWTime = o.TWTime
            this.TWEasing = o.TWEasing

            this.typewriterTagData = JSON.parse(o.typewriterTagData)
            this.TWData = JSON.parse(o.TWData)
            this.TWParamsOBJ = JSON.parse(o.TWParamsOBJ)
            this.TWParams = JSON.parse(o.TWParams)

            this.AnimFunctions = this.specialJSONparse(o.AnimFunctions)
            this.parsedText = this.specialJSONparse(o.parsedText)
            this._SFDXAliasFunctions = this.specialJSONparse(o._SFDXAliasFunctions)

            this.linkedDictionnaryUID = o.linkedDictionnaryUID
        }

        parseTypewriterParams(params)
        {
            var paramsA = params.trim().split(';');
            var defaultParams = {
                "value": {},
                "duration": {
                    "type": 5,
                    "fade": 10
                }
            }

            paramsA.forEach(param => {
                var paramA = param.trim().split(' ');
                let mode = paramA.shift().toLowerCase();
                let tag = paramA.shift().toLowerCase();
                let value = paramA.join();

                if (this.isANumber(value)) value = parseFloat(value)

                switch (mode)
                {
                    case "value":
                        defaultParams["value"][tag] = value
                        break;
                    case "duration":
                        switch (tag)
                        {
                            case "type":
                                defaultParams["duration"]["type"] = value
                                break;
                            case "fade":
                                defaultParams["duration"]["fade"] = value
                                break;

                            default:
                                console.warn(tag + " is not a recognised typewriter parameter for " + mode + ".")
                                break;
                        }
                        break
                    default:
                        console.warn(mode + " is not a recognised typewriter parameter.")
                        break;
                }
            });

            return defaultParams
        }

        getDefault(tag)
        {
            switch (tag)
            {
                case "offsetx":
                case "offsety":
                    return 0;
                case "opacity":
                    return 100;
                case "scale":
                case "scalex":
                case "scaley":
                    return 1;
                case "color":
                    return '#FFFFFF';
                default:
                    return 0;
            }
        }

        Tick2()
        {
            //Check if dictionnary is linked after load
            if (this.linkedDictionnaryUID != -1 && !this.linkedDictionnary)
            {
                this.linkedDictionnary = this._runtime.GetInstanceByUID(this.linkedDictionnaryUID)
            }

            //If instance action Set Text is called
            if (!this.SetTextCall && this.animated && this._inst._sdkInst._text != this.lastKnown)
            {
                this.animated = false;
            }

            if (this.SetTextCall) this.SetTextCall = false

            if (this.animated)
            {
                const time = this._runtime.GetGameTime();
                var str = ""
                var word = false
                if (this.typewriterActive)
                {
                    let id = 0
                    this.parsedText.forEach(el => {
                        for (let i = 0; i < el[1].length; i++)
                        {
                            let tags = {}

                            el[0].forEach(tag => {
                                if (typeof tag[1] === "function") tags[tag[0]] = tag[1](time, i)
                                else tags[tag[0]] = tag[1]
                            })

                            if (this.TWTime >= this.TWData.start[id][0] && this.TWData.data[id].hasOwnProperty("pause"))
                            {
                                this.typewriterPaused = true
                                delete this.TWData.data[id].pause;
                                this.Trigger(C3.Behaviors.skymen_Skymen_SpritefontDX.Cnds.OnTwPause)
                            }
                            if (this.TWTime >= this.TWData.start[id][0] && this.TWData.data[id].hasOwnProperty("fn"))
                            {
                                let fnData = this.TWData.data[id].fn.split(' ')
                                let fnName = fnData.shift()
                                let fnParams = JSON.parse("[" + fnData.join(' ') + "]") || []
                                c3_callFunction(fnName, fnParams)
                                delete this.TWData.data[id].fn;
                            }

                            Object.keys(this.TWData.data[id]).forEach(tag => {
                                if (!tags.hasOwnProperty(tag))
                                {
                                    tags[tag] = this.getDefault(tag)
                                }
                                tags[tag] = lerp(this.TWData.data[id][tag], tags[tag], EasingFunctions[this.TWEasing](unlerp(this.TWData.start[id][0], this.TWData.start[id][1], this.TWTime, true)))
                            })

                            let end = ""

                            Object.keys(tags).forEach(tag => {
                                if (!["wait", "fade", "type", "pause"].includes(tag))
                                {
                                    str += "[" + tag + "=" + tags[tag] + "]"
                                    end = "[/" + tag + "]" + end
                                }
                            })

                            let innerText = ""
                            let tmp = el[1][i]
                            if (typeof tmp != "string")
                            {
                                innerText += tmp()
                            }
                            else
                            {
                                innerText = tmp
                            }

                            str += innerText + end

                            if (this.TWTime >= this.TWData.start[id][0] && this.LastLetterID < id)
                            {
                                this.LastLetterID = id
                                this.Trigger(C3.Behaviors.skymen_Skymen_SpritefontDX.Cnds.OnLetterTyped)
                                word = true
                                this.curTypedWidth = str
                            }

                            if (word)
                            {
                                this.curTypedHeight = str
                                if (el[1][i] === " ") word = false
                            }

                            id++
                        }
                    })

                    if (this.TWTime >= this.TWData.start[this.TWData.start.length - 1][1])
                    {
                        //TW END
                        this.Trigger(C3.Behaviors.skymen_Skymen_SpritefontDX.Cnds.OnTwStop)
                        this.typewriterActive = false
                    }

                    if (!this.typewriterPaused) this.TWTime += this._runtime.GetDt(this._inst)
                }
                else
                {
                    this.parsedText.forEach(el => {
                        if (el[0].length == 0)
                        {
                            let innerText = ""
                            let tmp = el[1]
                            if (typeof tmp != "string")
                            {
                                tmp.forEach(f => {
                                    innerText += f()
                                })
                            }
                            else
                            {
                                innerText = tmp
                            }
                            //console.log(innerText)
                            str += innerText
                        }
                        else
                        {
                            for (let i = 0; i < el[1].length; i++)
                            {
                                let end = ""
                                el[0].forEach(tag => {
                                    let val
                                    if (typeof tag[1] === "function") val = tag[1](time, i)
                                    else val = tag[1]

                                    str += "[" + tag[0] + "=" + val + "]"
                                    end = "[/" + tag[0] + "]" + end
                                })
                                let innerText = ""
                                let tmp = el[1][i]
                                if (typeof tmp != "string")
                                {
                                    innerText += tmp()
                                }
                                else
                                {
                                    innerText = tmp
                                }
                                str += innerText + end
                            }
                        }
                    })
                }
                this.lastKnown = str
                //console.log(str)
                this._inst._sdkInst.CallAction(C3.Plugins.Spritefont2.Acts.SetText, str)
            }
        }

        Typewriter(text)
        {
            if (text === "")
            {
                this._inst._sdkInst.CallAction(C3.Plugins.Spritefont2.Acts.SetText, "");
                this.lastKnown = "";
                this.animated = false;
                this.typewriterActive = false
                return;
            }

            this.SetTextCall = true
            this.text = this.getTextWithNoTW(text)
            this.parseText()
            text = this.text
            this.LastLetterID = -1;
            this.typewriterActive = true
            this.animated = true
            this.typewriterPaused = false
            this.TWTime = 0;
            this.curTypedWidth = ""
            this.curTypedHeight = ""
            let pureText = this.getTextWithNoTags(this.text)
            let start = 0;
            this.TWData = {
                start: [],
                data: []
            }
            for (let i = 0; i < pureText.length; i++)
            {
                var curData = {}
                Object.assign(curData, this.TWParamsOBJ.value);
                Object.assign(curData, this.TWParamsOBJ.duration);
                for (let j = 0; j < this.typewriterTagData.length; j++)
                {
                    let el = this.typewriterTagData[j]
                    if (el.id > i) break;
                    let data = el.data.split(' ')
                    let tag = data[0].toLowerCase()
                    if ((tag === "wait" || tag === "pause" || tag === "fn") && el.id != i) continue;

                    if (tag === "fn")
                    {
                        data.shift()
                        curData[tag] = data.join(' ')
                    }
                    else if (this.isANumber(data[1])) curData[tag] = parseFloat(data[1])
                    else curData[tag] = data[1]
                }

                let special = false

                if (curData.hasOwnProperty("wait"))
                {
                    start += curData.wait
                }
                if (curData.hasOwnProperty("pause"))
                {
                    start += curData.fade
                }
                this.TWData.start.push([start, start + curData.fade])

                start += curData.type
                this.TWData.data.push(curData);
            }
            console.log(this.TWData)
            this.Trigger(C3.Behaviors.skymen_Skymen_SpritefontDX.Cnds.OnTwStart)
        }

        isAString(text)
        {
            var regex = /(["'])(\\?.)*?\1/g; //"
            var match = regex.exec(text);
            return match != null && match[0] === text;
        }

        isANumber(text)
        {
            return parseFloat(text).toString() === text;
        }

        SkipTypewriterToNextPause(toEnd)
        {
            if (!this.typewriterActive) return;

            let i = 0;
            while (this.TWTime > this.TWData.start[i][0])
            {
                i++
            }

            while (i < this.TWData.data.length && (toEnd || !this.TWData.data[i].hasOwnProperty("pause")))
            {
                if (this.TWData.data[i].hasOwnProperty("fn"))
                {
                    let fnData = this.TWData.data[id].fn.split(' ')
                    let fnName = fnData.shift()
                    let fnParams = JSON.parse("[" + fnData.join(' ') + "]") || []
                    c3_callFunction(fnName, fnParams)
                }
                i++
            }

            if (i === this.TWData.data.length)
            {
                this.Trigger(C3.Behaviors.skymen_Skymen_SpritefontDX.Cnds.OnTwStop)
                this.typewriterActive = false
            }
            else
            {
                this.TWTime = this.TWData.start[i][0]
            }
        }

        getTextWithNoTags(text)
        {
            let regex = /\[\/?((?!\/?tw|fn|text)[^\]]*)(=[^\]]*)?\]/gi;
            text = text.replace(regex, "");
            regex = /\[\/?((?!\/?tw)[^\]]*)(=[^\]]*)?\]/gi;
            var match;
            while ((match = regex.exec(text)) !== null)
            {
                let str = text
                let a = match[1].split('=')
                let b = a[1] || ""
                let c = b.split(' ')
                let len = parseInt(c[1]) || 0
                let str2 = ""
                for (let i = 0; i < len; i++)
                {
                    str2 += "0"
                }
                text = text.replace(match[0], str2)
                regex.lastIndex = 0;
            }
            return text;
        }

        getTextWithNoTW(text)
        {
            let regex = /\[tw=([^\]]*)?\]/gi;

            var match;
            var offset = 0
            this.typewriterTagData = []
            text = this.getVars(text)

            var noTag = this.getTextWithNoTags(text)

            while ((match = regex.exec(noTag)) !== null)
            {
                this.typewriterTagData.push(
                {
                    data: match[1].trim(),
                    id: match.index - offset
                })
                offset += match[0].length
            }
            return text.replace(regex, "");
        }

        getAnimFunction(tag)
        {
            if (!this.AnimFunctions.hasOwnProperty(tag))
            {
                let regex = /^[\d\w]+(\([^()]*\))?$/g;
                tag = tag.trim()
                if (regex.test(tag))
                {
                    var arr = tag.split('(')
                    var name = arr[0].trim().toLowerCase()
                    var fn = this._inst._objectType._SFDXAliasFunctions[name];

                    if (arr.length > 1)
                    {
                        var params = arr[1].slice(0, -1).split(',').map(function(s)
                        {
                            return s.trim()
                        });;
                        for (let i = 0; i < params.length; i++)
                        {
                            if (!isNaN(Number(params[i])))
                            {
                                params[i] = Number(params[i])
                            }
                        }
                        params.unshift(null)
                        this.AnimFunctions[tag] = fn.bind.apply(fn, params)
                    }
                    else
                    {
                        this.AnimFunctions[tag] = fn
                    }

                }
                else
                {
                    this.AnimFunctions[tag] = new Function("t", "i", "return " + tag.toLowerCase() + ";");
                }
            }
            return this.AnimFunctions[tag]
        }

        parseText()
        {
            var res = [];
            var stack = 0;
            var currentTag = []
            var currentText = ""
            var tagParam = false;
            var self = this;

            if (typeof this.linkedDictionnary !== "undefined")
            {
                this.replaceVars()
            }

            var text = this.text

            var regex = /\[\/?((?!\/?sfdx|text|fn)[^\]]*)\]/gi //Matches for tags that are neither sfdx nor text
            var match;
            while ((match = regex.exec(text)) !== null)
            {
                let str = text
                if (match[0].includes('='))
                {
                    let a = match[1].split('=')
                    text = text.replace(match[0], "[sfdx=" + a[0] + " \"" + a[1] + "\"]")
                }
                else
                {
                    text = text.replace(match[0], "[/sfdx]")
                }
                regex.lastIndex = 0;
            }

            this.text = text

            for (let i = 0; i < text.length; i++)
            {
                if (tagParam)
                {
                    if (text[i] === "]")
                    {
                        tagParam = false;
                        currentText = ""
                    }
                    else
                    {
                        currentTag[stack - 1] += text[i]
                    }
                }
                else
                {
                    if (text.substring(i).toLowerCase().startsWith("[sfdx="))
                    {
                        push(JSON.parse(JSON.stringify(currentTag)), currentText)
                        stack++;
                        currentTag.push("");
                        tagParam = true;
                        i += 5
                    }
                    else if (text.substring(i).toLowerCase().startsWith("[/sfdx]"))
                    {
                        stack--;
                        push(JSON.parse(JSON.stringify(currentTag)), currentText)
                        currentTag.pop();
                        currentText = ""
                        i += 6
                    }
                    else if (text.substring(i).toLowerCase().startsWith("[text="))
                    {
                        push(JSON.parse(JSON.stringify(currentTag)), currentText)
                        currentText = ""
                        i += 6
                        while (text[i] !== "]")
                        {
                            currentText += text[i]
                            i++
                        }
                        let textParam = currentText.split(' ')
                        let obj = []
                        let textName = textParam[0]
                        let length = parseInt(textParam[1]) || 0
                        for (let i = 0; i < length; i++)
                        {
                            obj.push(this.getChar.bind(this, textName, i, false))
                        }
                        push(JSON.parse(JSON.stringify(currentTag)), obj)
                        currentText = ""
                    }
                    else if (text.substring(i).toLowerCase().startsWith("[fn="))
                    {
                        push(JSON.parse(JSON.stringify(currentTag)), currentText)
                        currentText = ""
                        i += 4
                        while (text[i] !== "]")
                        {
                            currentText += text[i]
                            i++
                        }
                        let textParam = currentText.split(' ')
                        let obj = []
                        let textName = textParam[0]
                        let length = parseInt(textParam[1]) || 0
                        for (let i = 0; i < length; i++)
                        {
                            obj.push(this.getChar.bind(this, textName, i, true))
                        }
                        push(JSON.parse(JSON.stringify(currentTag)), obj)
                        currentText = ""
                    }
                    else
                    {
                        currentText += text[i]
                    }
                }
            }

            push(JSON.parse(JSON.stringify(currentTag)), currentText)

            function push(tag, text)
            {
                if (text == "") return

                var tagArray = []

                tag.forEach(t => {
                    var arr = t.split(' ')
                    var firstTag = arr.shift();
                    var tagfn = arr.join(' ');
                    var fn
                    if (self.isAString(tagfn))
                    {
                        fn = tagfn.substring(1, tagfn.length - 1)
                    }
                    else
                    {
                        fn = self.getAnimFunction(tagfn);
                    }

                    tagArray.push([firstTag, fn])
                })

                res.push([tagArray, text])
            }
            //console.log(res)
            this.parsedText = res
        }

        getChar(name, i, fn)
        {
            if (fn)
            {
                name = name.toLowerCase()
                let str = c3_callFunction(name) || "";
                if (i < str.length)
                {
                    return str[i]
                }
                else
                {
                    return " "
                }
            }
            else
            {
                name = name.toLowerCase()
                if (this.linkedDictionnary.has(name))
                {
                    let val = this.linkedDictionnary.get(name)
                    return i < val.length ? val[i] : " "
                }
                else
                {
                    return " "
                }
            }
        }

        replaceVars()
        {
            this.text = this.getVars(this.text)
        }

        getVars(text)
        {
            var regex = /\[(var|varfn)=([\d\w]+)\]/gi
            var match;
            while ((match = regex.exec(text)) !== null)
            {
                let isVar = match[1].trim().toLowerCase() === "var";
                let varName = match[2].toLowerCase();
                //console.log(match)
                if (isVar)
                {
                    if (this.linkedDictionnary.has(varName))
                    {
                        text = text.replace(match[0], this.linkedDictionnary.get(varName))
                        regex.lastIndex = 0;
                    }
                }
                else
                {
                    let str = c3_callFunction(varName) || ""
                    text = text.replace(match[0], str)
                    regex.lastIndex = 0;
                }
            }
            return text;
        }
    };
}